# ゲームをコントロールするプログラム

import bullet
import enemy
import player
import pygame as pg
import sound
import status


class Subject:  # 配信者（subject）の基本形
    # 初期化メソッド
    def __init__(self):
        # プロパティ
        self._observers = []  # 受信者（observer）を格納するためのリスト

    # メソッド
    def attach(self, observer):  # 受信者を登録する
        self._observers.append(observer)

    def notify(self, ntype):  # 受信者に通知する
        for observer in self._observers:
            # 引数「ntype」は、どのタイプの通知なのかを表す
            observer.update(ntype)


class GameManager(Subject):  # ゲーム管理用のクラス
    # 初期化メソッド
    def __init__(self, screen):
        super().__init__()
        # プロパティ
        # 主人公（Playerインスタンス）を作成
        self._player = player.Player()
        # 敵（Enemyインスタンス）を格納するための空のリストを作成
        self._enemies = []
        # 爆発エフェクト（BombEffectインスタンス）を格納するための空のリストを作成。一度に複数爆発することもあるため、リストに格納する。
        self._effects = []
        # 弾（Bullet）を格納するための空のリストを作成
        self._bullets = []
        # 敵工場インスタンス（ファクトリーインスタンス）を作成
        self._factory = enemy.EnemyFactory()
        # 受信者インスタンス（オブザーバーインスタンス）を作成
        self._status = status.Status()  # 上部のステータス情報用のプロパティ
        self.attach(self._status)  # 受信者を登録する
        # プロパティをGameManagerインスタンスのreset()メソッドで初期化する
        self.reset(screen)

    @property  # _is_playingプロパティに対するゲッターメソッド
    def is_playing(self):
        return self._is_playing

    @property  # _is_clearedプロパティに対するゲッターメソッド
    def is_cleared(self):
        return self._is_cleared

    # メソッド
    def reset(self, screen):  # ゲームのリセット（初期化）
        # 今はゲーム中か、ゲームが停止中かどうかを知らせるためのフラグとなるプロパティ。通常はTrue、停止中はFalseにする。
        self._is_playing = True
        # ゲームクリアしたかどうかのフラグとなるプロパティ
        self._is_cleared = False
        self._player.reset()  # Playerインスタンスのreset()メソッド
        self._enemies.clear()  # リストのclear()メソッドで敵リストを空にする
        # 敵を定期的に作成させるためのカウントプロパティ
        self._spawn_count = 0
        self._bullets.clear()  # 弾リストを空に
        # 弾を一定間隔で発射するためのカウントプロパティ
        self._bullet_count = 0
        self._status.reset()  # スコアと進んだ距離をリセット（初期化）
        # SoundManagerインスタンスのbgmstart()メソッドでBGMをループ再生する
        sound.SoundManager.get_instance().bgmstart()
        """
        for i in range(2):  # 通常の敵を２体作成
            self._enemies.append(enemy.Enemy(screen))
        for i in range(1):  # 炎の敵を１体作成
            self._enemies.append(enemy.FlameEnemy(screen))
        for i in range(1):  # 氷の敵を１体作成
            self._enemies.append(enemy.IceEnemy(screen))
        """

    def update(self, screen):  # 更新処理
        self.notify("distance")  # 進んだ距離を受信者に通知する
        self._bullet_count += 1
        if self._bullet_count > 10:  # 弾の発生量（最短１０カウントに１回）
            key = pg.key.get_pressed()
            if key[pg.K_a]:  # Aキーで弾を発射
                self._bullets.append(bullet.Bullet(self._player.rect))
                self._bullet_count = 0
        for e in self._effects:
            e.update()  # 爆発エフェクトリストの爆発エフェクトを更新
        for b in self._bullets:
            b.update()  # 弾リストの弾を更新
        self._player.update(screen)  # 主人公の更新
        self._spawn_count += 1
        # カウンターが１５になったら次の敵を生成し、敵を定期的に出現させるようにする（敵の発生量）
        if self._spawn_count > 15:
            self._spawn_count = 0
            self._enemies.append(self._factory.random_create(screen))
        for e in self._enemies:
            for b in self._bullets:
                # 発射中の弾と敵が衝突していたら
                if e.rect.colliderect(b.rect):
                    # 主人公の攻撃音を再生する
                    sound.SoundManager.get_instance().playattack()
                    self._bullets.remove(b)  # 弾リストから発射中の弾を削除
                    e.hp -= 50  # 敵のHPを少し減らす
                    if e.hp <= 0:  # 敵のHPが０になって倒した場合
                        self.notify("score")  # スコアを受信者に通知する
                        # 引数e.rectは、敵（Enemyインスタンス）の配置位置（Rectオブジェクト）
                        bomb = enemy.BombEffect(e.rect, self._effects)
                        # 敵を破壊した時の音を再生する
                        sound.SoundManager.get_instance().playblast()
                        self._effects.append(bomb)
                        self._enemies.remove(e)
                        if self._status.score == 30:  # クリア条件
                            self._is_playing = False  # ゲーム停止
                            self._is_cleared = True  # ゲームクリア
                            # BGMを停止する
                            sound.SoundManager.get_instance().bgmstop()
                            # ゲームクリア音を再生する
                            sound.SoundManager.get_instance().playclear()
            # 敵が死んでいたら敵のリストから削除
            if e.is_alive == False:
                self._enemies.remove(e)
                break
            e.update(screen)  # 敵の更新
            if e.rect.y >= screen.get_height():
                self._enemies.remove(e)  # 画面外まで落下した敵は消える
            # 敵と主人公が衝突したら、敵を上に移動し、敵の向きを変更する
            # pygame.Rect.colliderect(Rectオブジェクト)メソッドは、２つのRectオブジェクトが重なっているかどうか調べる。２つのRectオブジェクトの一部分が重なっている場合にはTrueが返される。(重なっているのが上辺と下辺、もしくは、右辺と左辺の場合は重なっているとは見なされない)
            # 弾で消したはずの敵と衝突しないように、敵が_enemiesリストプロパティの中に存在する時だけ敵と主人公の衝突判定を行う。敵と主人公が衝突すると主人公はダメージを受ける。
            if e in self._enemies:
                if e.rect.colliderect(self._player.rect):
                    # 主人公がダメージを受けた時の音を再生する
                    sound.SoundManager.get_instance().playbomb()
                    self._enemies.remove(e)  # 衝突した敵は敵リストから削除
                    # Playerインスタンスのdamage()メソッド
                    self._player.damage()
                    self._player.hp -= 50
                    if self._player.hp <= 0:
                        self._is_playing = False
                        # BGMを停止する
                        sound.SoundManager.get_instance().bgmstop()
                        # ゲームオーバー音を再生する
                        sound.SoundManager.get_instance().playover()

    def draw(self, screen):  # 描画処理
        for b in self._bullets:  # 弾の描画
            b.draw(screen)
        for e in self._effects:  # 爆発エフェクトの描画
            e.draw(screen)
        self._player.draw(screen)  # 主人公の描画
        for e in self._enemies:
            e.draw(screen)  # 敵の描画
        self._status.draw(screen)  # 画面上部のステータス情報の描画
